import base64
from abc import ABC, abstractmethod
import logging
from typing import Tuple

import boto3
from pulumi import automation as auto
from pulumi.automation import LocalWorkspaceOptions, ProjectBackend, ProjectSettings

from blazetest.core.config import CWD, LOKI_HOST, LOKI_USER, DOCKER_FILE_PATH
from blazetest.core.deployment.aws import AWSWorkflow
from blazetest.core.image_build.image_build import ImageBuildPush

logger = logging.getLogger(__name__)


class InfraSetupTool(ABC):
    def __init__(
        self,
        aws_region: str,
        resource_prefix: str,
        s3_bucket_name: str,
        ecr_repository_prefix: str,
        lambda_function_timeout: int,
        lambda_function_memory_size: int,
        loki_api_key: str,
        build_backend: str,
        depot_token: str,
        depot_project_id: str,
        tags: dict,
        debug: bool,
    ):
        self.aws_region = aws_region
        self.resource_prefix = resource_prefix
        self.s3_bucket_name = s3_bucket_name
        self.ecr_repository_prefix = ecr_repository_prefix
        self.lambda_function_timeout = lambda_function_timeout
        self.lambda_function_memory_size = lambda_function_memory_size
        self.loki_api_key = loki_api_key
        self.depot_token = depot_token
        self.depot_project_id = depot_project_id
        self.tags = tags
        self.debug = debug
        self.build_backend = build_backend

    @abstractmethod
    def deploy(self) -> None:
        pass


def log_pulumi_event(event: str):
    logger.info(event)


class PulumiInfraSetup(InfraSetupTool):
    """
    Uses Pulumi (https://www.pulumi.com/docs/) Automation API to build and deploy artifacts to the cloud.
    """

    PROJECT_NAME = "blazetest"
    PROJECT_BACKEND_URL = "file:\\/\\/~"
    PROJECT_RUNTIME = "python"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def deploy(self) -> None:
        env_vars = {}

        repo_info = self.create_ecr_repository()
        image_uri = self.build_and_push_image(
            repo_info=repo_info, build_backend=self.build_backend
        )

        workflow = AWSWorkflow(
            resource_prefix=self.resource_prefix,
            s3_bucket_name=self.s3_bucket_name,
            image_uri=image_uri,
            lambda_function_timeout=self.lambda_function_timeout,
            lambda_function_memory_size=self.lambda_function_memory_size,
            loki_host=LOKI_HOST,
            loki_user=LOKI_USER,
            loki_api_key=self.loki_api_key,
            env_vars=env_vars,
            tags=self.tags,
        )

        stack = auto.create_stack(
            stack_name=self.resource_prefix,
            project_name="blazetest",
            program=workflow.deploy,
            opts=self.get_project_settings(),
        )

        logger.info("Installing plugins")

        # TODO: updated automatically to the stable version
        stack.workspace.install_plugin("aws", "v5.42.0")
        stack.workspace.install_plugin("docker", "v4.3.1")

        stack.set_config("aws:region", auto.ConfigValue(value=self.aws_region))
        stack.refresh(on_output=log_pulumi_event, show_secrets=False)

        logger.info("Deploying..")
        workflow_result = stack.up(  # noqa
            show_secrets=False, on_output=log_pulumi_event, debug=self.debug
        )

        logger.info(
            "Deploying has finished.",
        )

    def __get_ecr_login_token(self) -> Tuple[str, str]:
        ecr = boto3.client("ecr", region_name=self.aws_region)

        response = ecr.get_authorization_token()

        auth_data = response["authorizationData"][0]
        token = auth_data["authorizationToken"]
        decoded_token = base64.b64decode(token).decode("utf-8")
        username, password = decoded_token.split(":")

        return username, password

    def create_ecr_repository(self):
        ecr = boto3.client("ecr", region_name=self.aws_region)

        ecr_tags = []
        for tag in self.tags:
            ecr_tags.append({"Key": tag, "Value": self.tags[tag]})

        try:
            response = ecr.create_repository(
                repositoryName=self.ecr_repository_prefix, tags=ecr_tags
            )
            repository = response["repository"]
            logger.info(
                f"Repository '{self.ecr_repository_prefix}' created successfully."
            )
            return repository
        except ecr.exceptions.RepositoryAlreadyExistsException:
            logger.warning(f"Repository '{self.ecr_repository_prefix}' already exists.")
        except Exception as e:
            logger.error(f"An error occurred: {e}")

    def build_and_push_image(self, repo_info: dict, build_backend: str = "depot"):
        image_uri = f"{repo_info['repositoryUri']}:{self.resource_prefix}"

        image_build_push = ImageBuildPush(
            backend=build_backend,
            project_context=CWD,
            docker_file_path=DOCKER_FILE_PATH,
            image_uri=image_uri,
            build_platform="linux/amd64",
        )

        image_build_push.build()

        # TODO: check if it works in the CI/CD pipeline
        username, password = self.__get_ecr_login_token()
        image_build_push.login(
            username=username,
            password=password,
            registry=repo_info["repositoryUri"],
        )

        image_build_push.push()

        return image_uri

    def get_project_settings(self):
        return LocalWorkspaceOptions(
            project_settings=ProjectSettings(
                name=self.PROJECT_NAME,
                backend=ProjectBackend(url=self.PROJECT_BACKEND_URL),
                runtime=self.PROJECT_RUNTIME,
            ),
        )



class Repository:
    def __init__(self, name: str, tags: dict):
        self.name = name
        self.tags = tags


class ECRRepository(Repository):
    def __init__(self, name: str, tags: dict, registry_id: str):
        super().__init__(name, tags)
        self.registry_id = registry_id



import os
from abc import ABC

import docker

from blazetest.core.utils.command_executor import CommandExecutor
from blazetest.core.utils.exceptions import DepotTokenNotProvided


class BuildTool(ABC):
    EXECUTABLE = None

    def __init__(
        self,
        project_context: str,
        docker_file_path: str,
        image_uri: str,
        build_platform: str,
        *args,
        **kwargs,
    ):
        self.project_context = project_context
        self.docker_file_path = docker_file_path
        self.image_uri = image_uri
        self.build_platform = build_platform

    def login(self, username: str, password: str, registry: str):
        pass

    def build(self):
        pass

    def push(self):
        pass

    def build_and_push(self):
        pass

    def execute(self, command: str, arguments: dict, allowed_return_codes=None) -> int:
        if allowed_return_codes is None:
            allowed_return_codes = [0]

        command_executor = CommandExecutor(
            executable=self.EXECUTABLE,
            command=command,
            arguments=arguments,
        )
        command_result = command_executor.execute_command(
            allowed_return_codes=allowed_return_codes
        )
        return command_result


class DepotBuildTool(BuildTool):
    """
    Uses depot.dev to build and push images to a remote repository.
    """

    EXECUTABLE = "depot"  # TODO: would executable work correctly?
    BUILD_COMMAND = "build"

    def __init__(
        self,
        project_context: str,
        docker_file_path: str,
        image_uri: str,
        build_platform: str,
        depot_token: str = None,
        depot_project_id: str = None,
    ):
        super().__init__(
            project_context=project_context,
            docker_file_path=docker_file_path,
            image_uri=image_uri,
            build_platform=build_platform,
        )
        self.depot_token = depot_token
        if self.depot_token is None:
            self.depot_token = os.getenv("DEPOT_TOKEN")
            if self.depot_token is None:
                raise DepotTokenNotProvided(
                    "Depot token not provided. "
                    "Please provide it using --depot-token CLI argument or DEPOT_TOKEN environment variable."
                )
        self.depot_project_id = depot_project_id

    def build_and_push(self):
        args = {
            "--tag": self.image_uri,
            "--file": self.docker_file_path,
            "--platform": self.build_platform,
            "--token": self.depot_token,
            "--push": None,
            "--provenance": "false",
            self.project_context: None,
        }

        if self.depot_project_id:
            args["--project"] = self.depot_project_id

        return self.execute(
            command=self.BUILD_COMMAND,
            arguments=args,
        )

    def login(self, username: str, password: str, registry: str = None):
        client = docker.from_env()
        client.login(username=username, password=password, registry=registry)


class DockerBuildTool(BuildTool):
    """
    This class will be used to build and push images from local Docker.
    """

    EXECUTABLE = "docker"
    BUILD_COMMAND = "build"
    PUSH_COMMAND = "push"

    def __init__(
        self,
        project_context: str,
        docker_file_path: str,
        image_uri: str,
        build_platform: str,
    ):
        super().__init__(
            project_context=project_context,
            docker_file_path=docker_file_path,
            image_uri=image_uri,
            build_platform=build_platform,
        )

    def login(self, username: str, password: str, registry: str):
        client = docker.from_env()
        client.login(username=username, password=password, registry=registry)

    def build_and_push(self):
        self.build()
        self.push()

    def build(self):
        args = {
            "--tag": self.image_uri,
            "--file": self.docker_file_path,
            "--platform": self.build_platform,
            self.project_context: None,
        }

        return self.execute(
            command=self.BUILD_COMMAND,
            arguments=args,
        )

    def push(self):
        args = {
            self.image_uri: None,
        }

        return self.execute(
            command=self.PUSH_COMMAND,
            arguments=args,
        )


class DockerCloudBuildTool(BuildTool):
    """
    This class will be used to build and push images from Docker Cloud.
    """

    EXECUTABLE = "docker buildx"
    CREATE_COMMAND = "create"
    BUILD_COMMAND = "build"
    PUSH_COMMAND = "push"

    def __init__(
        self,
        project_context: str,
        docker_file_path: str,
        image_uri: str,
        build_platform: str,
    ):
        super().__init__(
            project_context=project_context,
            docker_file_path=docker_file_path,
            image_uri=image_uri,
            build_platform=build_platform,
        )

    def login(self, username: str, password: str, registry: str):
        client = docker.from_env()
        client.login(username=username, password=password, registry=registry)

    def build_and_push(self):  # aws_username: str, aws_password: str, aws_registry: str
        self.build()
        # Login to ECR
        # self.login()
        self.push()

    def create_cloud_builder(self):
        args = {
            "--driver": "cloud",
            "--name": "builder",
        }

        return self.execute(
            command=self.CREATE_COMMAND,
            arguments=args,
        )

    def build(self):
        args = {
            "--tag": self.image_uri,
            "--file": self.docker_file_path,
            "--platform": self.build_platform,
            self.project_context: None,
        }

        return self.execute(
            command=self.BUILD_COMMAND,
            arguments=args,
        )

    def push(self):
        args = {
            self.image_uri: None,
        }

        return self.execute(
            command=self.PUSH_COMMAND,
            arguments=args,
        )


class RemoteBuildTool(BuildTool):
    """
    This class will be used to build and push images from AWS CodeBuild, Google Cloud Build, etc.
    """

    def __init__(self):
        raise NotImplementedError("RemoteBuildTool is not implemented yet.")

    def build(self):
        pass

    def push(self):
        pass
